.fretboard-chord-diagram-top-toggle, .fretboard-chord-diagram-top-toggle image {
  color: rgb(0, 0, 0);
}

.fretboard-chord-diagram-top-toggle.open, .fretboard-chord-diagram-top-toggle image.open {
  color: @accent_bg_color;
}

.fretboard-chord-diagram-top-toggle button:checked:hover {
  box-shadow: 0px 0px 0px 1px rgba(0, 0, 0, 0.5);
}

.fretboard-chord-name-entry entry placeholder {
  color: rgba(0, 0, 0, 0.6);
}

.fretboard-star-toggle:checked:not(:hover) {
  border: 0;
  box-shadow: none;
}

.fretboard-chord-preview-dot.active {
  border: 1px solid @border-color;
}
