.debug {
  outline: 2px solid green;
}

barre-spin button {
  border-radius: 50%;
  background: transparent;
}

barre-spin button:hover {
  background: rgba(153, 153, 153, 0.3);
}

barre-spin label {
  font-size: 2rem;
  font-weight: bold;
}

chord-diagram-toggle button, .fretboard-chord-diagram-top-toggle button{
  border-radius: 50%;
  padding: 0;
  transition-duration: 150ms;
}

chord-diagram-toggle button:checked {
  background: var(--accent-bg-color);
}

chord-diagram-toggle button:hover {
  background: rgba(153, 153, 153, 0.3);
}

chord-diagram-toggle button:checked:hover {
  background: var(--accent-bg-color);
}

.fretboard-chord-diagram-top-toggle button:checked{
  background: transparent;
  box-shadow: none;
}

.fretboard-chord-diagram-top-toggle, .fretboard-chord-diagram-top-toggle image {
  color: rgba(153, 153, 153, 0.4);
}

.fretboard-chord-diagram-top-toggle.open, .fretboard-chord-diagram-top-toggle image.open {
  color: var(--accent-bg-color);
}

.fretboard-chord-diagram-top-toggle button:hover {
  background: rgba(153, 153, 153, 0.3);
}

.fretboard-chord-name-entry entry {
  font-size: 1.8rem;
  padding: 8px 6px;
  font-weight: 700;
  border-radius: 12px;
}

.fretboard-chord-name-entry entry placeholder {
  color: rgba(125, 125, 125, 0.6);
}

.fretboard-star-toggle:checked {
  background: transparent;
}

.fretboard-star-toggle:checked:hover {
  background: rgba(125, 125, 125, 0.15);
}

.fretboard-chord-preview-dot {
  background: transparent;
  border-radius: 50%;
}

.fretboard-chord-preview-dot.active {
  background: var(--accent-bg-color);
} 

.fretboard-chord-preview-button {
  padding: 16px 0;
  border-radius: 12px;
}

.fretboard-chord-preview-button:hover {
  background-color: alpha(currentColor, 0.025);
}
