.fretboard-chord-diagram-top-toggle, .fretboard-chord-diagram-top-toggle image {
  color: rgb(255, 255, 255);
}

.fretboard-chord-diagram-top-toggle.open, .fretboard-chord-diagram-top-toggle image.open {
  color: var(--accent-bg-color);
}

.fretboard-chord-diagram-top-toggle button:checked:hover {
  box-shadow: 0px 0px 0px 1px rgba(255, 255, 255, 0.6);
}

.fretboard-chord-name-entry entry placeholder {
  color: rgba(255, 255, 255, 0.6);
}
